/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _XOPEN_SOURCE 500
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

int store_size(int fd)
{
  struct stat stbuf;
  fstat(fd,&stbuf);
  if (stbuf.st_size % 512) return 0;
  return stbuf.st_size/512;
}

uint32_t hash_sector(uint8_t *data, int hash_size)
{
  uint32_t r=0;
  int i;
  for (i=0;i<512;++i)
    r=((r<<8)+data[i])%hash_size;
  //printf("hash=%d\n",r);
  return r;
}

uint8_t idxidx[128];
uint32_t indexes[128];

int idxidxcmp(uint8_t *a, uint8_t *b)
{
  uint32_t v1=indexes[*a];
  uint32_t v2=indexes[*b];
  return 2*(v1>v2)-1;
}

int prime(int p)
{
  int f;
  if (p<4) return 1;
  if (!(p&1)) return 0;
  for (f=3;f*f<=p;f+=2)
    if (!(p%f)) return 0;
  return 1;
}

int main(int argc, char ** argv)
{
  uint32_t hash_size;
  uint32_t size;
  int store_fd,hash_fd;
  uint8_t buf[512];
  uint8_t buf2[512];
  char *namebuf;
  int r;
  uint32_t i;
  int maxused=0;

  if (argc != 2) {
    fprintf(stderr,"Uage: %s <store>\n",argv[0]);
    return 1;
  }

  store_fd=open(argv[1],O_RDONLY);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  size=store_size(store_fd);
  if (size) {
    switch (read(store_fd,buf,512)) {
    case -1:
      perror(argv[1]);
      return 1;
    default:
      size=0;
    case 512:
      if (memcmp(buf,"TSTORE1",7)) size=0;
    }
  }
  if (!size) {
    fprintf(stderr,"%s: Not a valid store",argv[1]);
    return 1;
  }

  hash_size=(size+65)/64;
  while (!prime(hash_size)) ++hash_size;

  namebuf=malloc(strlen(argv[1])+10);
  if (!namebuf) {
    perror("Out of memory");
    return 1;
  }

  sprintf(namebuf,"%s.idx",argv[1]);

  hash_fd=open(namebuf,O_CREAT|O_RDWR|O_TRUNC,0666);
  if (hash_fd==-1) {
    perror(namebuf);
    return 1;
  }
  ftruncate(hash_fd,hash_size*512);

  for (i=1;(r=read(store_fd,buf,512))==512;++i) {
    uint32_t this_hash=hash_sector(buf,hash_size);
    int used,j,min,max;
    //printf("this: %d\n",this_hash);
    if (pread(hash_fd,indexes,512,this_hash*512)!=512) {
      perror(namebuf);
      return 1;
    }

    for (used=0;(used<128)&&(indexes[used]);++used)
      idxidx[used]=used;
    if (used>maxused) maxused=used;
    if (used==128) {
      fprintf(stderr,"Hash overflow\n");
      return 1;
    }
    if (used>1)
      qsort(idxidx,used,1,(void*)idxidxcmp);
    min=0;
    max=used;
    for (j=0;min<max;++j) {
      int midt=idxidx[j];
      int cmpr;
      //printf("%d,%d,%d\n",min,midt,max);
      if ((min<=midt)&&(midt<max)) {
	//printf("i %d\n",indexes[midt]);
	switch (pread(store_fd,buf2,512,indexes[midt]*512)) {
	case -1:
	  perror(argv[1]);
	  return 1;
	default:
	  fprintf(stderr,"%s: Incomplete pread\n",argv[1]);
	  return 1;
	case 512:
	  break;
	}
	//printf(">>%d\n",this_hash);
	cmpr=memcmp(buf2,buf,512);
	//printf("%d\n",cmpr);
	assert(hash_sector(buf2,hash_size)==this_hash);
	if (cmpr==0) {
	  fprintf(stderr,"Duplicate sector\n");
	  return 1;
	} else {
	  if (cmpr>0) {
	    max=midt;
	  } else {
	    min=midt+1;
	  }
	} /* if cmpr */
      } /* if min<=midt<max */
    } /* for */
    assert((0<=min)&&(min==max)&&(max<=used)&&(used<128));
    memmove(indexes+min+1,indexes+min,(used-min)*4);
    indexes[min]=i;
    if (pwrite(hash_fd,indexes,512,this_hash*512)!=512) {
      perror(namebuf);
      return 1;
    }
  }

  switch(r) {
  default:
    fprintf(stderr,"%s: Incomplete read\n",argv[1]);
    return 1;
  case -1:
    perror(argv[1]);
    return 1;
  case 0:
    break;
  }
  if (i != size) {
    fprintf(stderr,"%s: Has %s as we read it\n",
	    argv[1],(i>size)?"grown":"shrunk");
    return 1;
  }

  if (maxused>31) {
    printf("Largest hash %d of 128 used\n",maxused);
    if (maxused>95)
      printf("Rehashing suggested\n");
  }
  return 0;
}
