/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _XOPEN_SOURCE 500
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

int size_in_blocks(int fd)
{
  struct stat stbuf;
  fstat(fd,&stbuf);
  if (stbuf.st_size % 512) return 0;
  return stbuf.st_size/512;
}

uint32_t hash_sector(uint8_t *data, int hash_size)
{
  uint32_t r=0;
  int i;
  for (i=0;i<512;++i)
    r=((r<<8)+data[i])%hash_size;
  //printf("hash=%d\n",r);
  return r;
}

uint8_t idxidx[128];
uint32_t indexes[128];

int idxidxcmp(uint8_t *a, uint8_t *b)
{
  uint32_t v1=indexes[*a];
  uint32_t v2=indexes[*b];
  return 2*(v1>v2)-1;
}

ssize_t myread(int fd, void *buf, size_t count)
{
  ssize_t r,total=0;
  memset(buf,0,count);
  while (count) {
    r=read(fd,buf,count);
    if (r<1) return total?total:r;
    total+=r;
    buf=(((char*)buf)+r);
    count-=r;
  }
  return total;
}

uint32_t hash_size,store_size;
int store_fd,hash_fd;
char *namebuf;
char *storename;
int maxused=0;

int zblock(uint8_t *buf)
{
  int i;
  for (i=0;i<512;++i)
    if (buf[i]) return 0;
  return 1;
}

uint32_t put_block(uint8_t *buf)
{
  uint8_t buf2[512];
  uint32_t this_hash=hash_sector(buf,hash_size);
  int used,j,min,max;

  if (zblock(buf)) return 0;

  if (pread(hash_fd,indexes,512,this_hash*512)!=512) {
    perror(namebuf);
    exit(1);
  }

  for (used=0;(used<128)&&(indexes[used]);++used)
    idxidx[used]=used;
  if (used>maxused) maxused=used;
  if (used>1)
    qsort(idxidx,used,1,(void*)idxidxcmp);
  min=0;
  max=used;
  for (j=0;min<max;++j) {
    int midt=idxidx[j];
    int cmpr;
    //printf("%d,%d,%d\n",min,midt,max);
    if ((min<=midt)&&(midt<max)) {
      //printf("i %d\n",indexes[midt]);
      switch (pread(store_fd,buf2,512,indexes[midt]*512)) {
      case -1:
	perror(storename);
	exit(1);
      default:
	fprintf(stderr,"%s: Incomplete pread\n",storename);
	exit(1);
      case 512:
	break;
      }
      //printf(">>%d\n",this_hash);
      cmpr=memcmp(buf2,buf,512);
      //printf("%d\n",cmpr);
      assert(hash_sector(buf2,hash_size)==this_hash);
      if (cmpr==0) {
	/* Found a match */
	return indexes[midt];
      } else {
	if (cmpr>0) {
	  max=midt;
	} else {
	  min=midt+1;
	}
      } /* if cmpr */
    } /* if min<=midt<max */
  } /* for */
  assert((0<=min)&&(min==max)&&(max<=used)&&(used<=128));
  if (used==128) {
    fprintf(stderr,"Hash overflow\n");
    exit(1);
  }
  memmove(indexes+min+1,indexes+min,(used-min)*4);
  assert(write(store_fd,buf,512)==512);
  indexes[min]=store_size;
  if (pwrite(hash_fd,indexes,512,this_hash*512)!=512) {
    perror(namebuf);
    exit(1);
  }
  return store_size++;
}

uint32_t pointers[5][128];
int usage[5]={0,0,0,0};

void put_pointer(uint32_t p, int level)
{
  pointers[level][usage[level]]=p;
  usage[level]++;
  if (usage[level]==128) {
    put_pointer(put_block((void*)(pointers[level])),level+1);
    usage[level]=0;
  }
}

uint32_t flush_pointers()
{
  int i;
  for (i=0;i<4;++i)
    while(usage[i])
      put_pointer(0,i);
  return pointers[4][0];
}

int main(int argc, char ** argv)
{
  uint64_t filesize;
  uint32_t original_size;
  uint8_t buf[512];
  int r;

  storename=argv[1];

  if (argc != 3) {
    fprintf(stderr,"Uage: %s <store> <file>\n",argv[0]);
    return 1;
  }

  if (strlen(argv[2])>495) {
    fprintf(stderr,"%s: Name too long\n",argv[2]);
    return 1;
  }

  store_fd=open(argv[1],O_RDWR|O_APPEND);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  store_size=size_in_blocks(store_fd);
  if (store_size) {
    switch (pread(store_fd,buf,512,0)) {
    case -1:
      perror(argv[1]);
      return 1;
    default:
      store_size=0;
    case 512:
      if (memcmp(buf,"TSTORE1",7)) store_size=0;
    }
  }
  if (!store_size) {
    fprintf(stderr,"%s: Not a valid store\n",argv[1]);
    return 1;
  }

  original_size=store_size;

  namebuf=malloc(strlen(argv[1])+10);
  if (!namebuf) {
    perror("Out of memory");
    return 1;
  }

  sprintf(namebuf,"%s.idx",argv[1]);

  hash_fd=open(namebuf,O_RDWR,0666);
  if (hash_fd==-1) {
    perror(namebuf);
    return 1;
  }

  hash_size=size_in_blocks(hash_fd);
  if (!hash_size) {
    fprintf(stderr,"Invalid hash index (rebuild?)\n");
    return 1;
  }

  do {
    r=myread(0,buf,512);
    if (r==-1) {
      perror(argv[2]);
      return 1;
    }
    if (r) {
      filesize += r;
      put_pointer(put_block(buf),0);
    }
  } while(r==512);

  if (filesize) {
    uint32_t block_directory=flush_pointers();

    --original_size;
    memset(buf,0,512);
    memcpy(buf,&original_size,4);
    memcpy(buf+4,&block_directory,4);
    memcpy(buf+8,&filesize,8);
    strcpy(buf+16,argv[2]);

    while(1) {
      uint32_t direntry=put_block(buf);
      if(direntry+1 == store_size) break;
      //printf("%d,%d\n",prevsize,direntry);
      memset(buf,0,512);
      memcpy(buf,&direntry,4);
    }
  }

  if (maxused>31) {
    printf("Largest hash %d of 128 used\n",maxused);
    if (maxused>95)
      printf("Rehashing suggested\n");
  }

  return 0;
}
